/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
using JS;
using Nuvola.JSTools;
using Diorite.Logger;

namespace Nuvola{

/**
 * Errors thrown from Nuvola Palyer JavaScript API
 */
public errordomain JSError{
	/**
	 * An object has not been found
	 */
	NOT_FOUND,
	/**
	 * A value has wrong type
	 */
	WRONG_TYPE,
	/**
	 * Call of a JavaScript function failed.
	 */
	FUNC_FAILED,
	/**
	 * Unable to load script from file
	 */
	READ_ERROR,
	/**
	 * JavaScript API does not have any context yet.
	 */
	NO_CONTEXT,
	/**
	 * Execution of a script caused an exception.
	 */
	EXCEPTION;
}


/**
 * Nuvola JavaScript API provides interface for service integrations to communicate
 * with Nuvola Player runtime. Practically, it is a bridge between Service object
 * and JavaScript environment of WebKit WebView.
 * 
 * Main method of the main Nuvola Player JavaScript object are implemented here,
 * other helper functions and tools are loaded from a JavaScript file.
 */
public class JSApi: GLib.Object{
	/**
	 * Name of file with integration script.
	 */
	private static const string INTEGRATION_FILENAME = "integration.js";
	/**
	 * Name of file with settings script.
	 */
	private static const string SETTINGS_SCRIPT = "settings.js";
	/**
	 * Major version of the JavaScript API
	 */
	public static const int API_VERSION = 2;
	public static const int API_VERSION_MINOR = 1;
	private static const int COMPAT_API_MAJOR = 1;
	private static const string MESSAGE_RECEIVER = "onMessageReceived";
	private static const string MAIN_SCRIPT_FILENAME = "js/main.js";
	private static const string COMPAT_SCRIPT_FILENAME = "js/compat-1.x.js";
	public static const string CONFIG_CHANGED_MESSAGE = "config-changed";
	private static const string CONFIG_OBJECT = "config";
	private static unowned JS.Class klass;
	/**
	 * Identifier of the main frame
	 */
	public static const string MAIN_FRAME_ID = "__main__";
	/**
	 * Identifier of the frame for service's preferences.
	 */
	public static const string PREFERENCES_FRAME_ID = "__preferences__";
	
	
	private SList<weak WebKit.WebFrame> frames;
	public Service service {get; private set;}
	private Diorite.Actions? actions;
	private Nuvola.WebView view;
	private Diorite.Storage storage;
	private HashTable<string, bool?> changed_actions;
	private bool alternate_injection = false;
	private File integration_file;
	private static const string FRAME_PROPERTY_NAME = "nuvola_frame_bridge";
	
	[Description(nick = "Main frame bridge", blurb = "Main frame bridge associated with a toplevel frame.")]
	public FrameBridge main_bridge{ get; private set;}
	
	/**
	 * Emitted when configuration of a service has been changed from a script.
	 */
	public signal void config_changed();
	
	/**
	* Emitted when Nuvola Player JavaScript API is loaded in the main frame.
	*/
	public signal void load_finished();
	
	/**
	* Emitted when Nuvola Player JavaScript API has been injected.
	* 
	* @param bridge    bridge used to inject API
	*/
	public signal void api_injected(FrameBridge bridge);

	/**
	 * Emitted by script when song details have been changed.
	 * @param song				song name
	 * @param artist			artist name
	 * @param album				album name
	 * @param art_uri			URI of album art
	 * @param playback_state	state of playback
	 */
	public signal void song_changed(string? song, string? artist, string? album, string? art_uri, string playback_state);
	
	/**
	 * Emitted when new bridge is created.
	 * 
	 * @param bridge    new bridge
	 */
	public signal void bridge_created(FrameBridge bridge);
		
	/**
	 * Updates information about a current song and state. Implementation of a method Nuvola.updateSong()
	 */
	static unowned JS.Value update_song_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception){
		unowned JS.Value undefined = JS.Value.undefined(ctx);
		exception = null;
		
		/* Check number of arguments */
		if(args.length != 5){
			critical("Nuvola.updateSong(): Wrong number of arguments: 5 required, %d provided", args.length);
			ctx.collect_garbage();
			return undefined;
		}
		
		/* Check type of arguments */
		// song
		if(!args[0].is_string(ctx) && !args[0].is_null(ctx)){
			critical("Nuvola.updateSong(): Arg 0 must be a string or null");
			ctx.collect_garbage();
			return undefined;
		}
		// artist
		if(!args[1].is_string(ctx) && !args[1].is_null(ctx)){
			critical("Nuvola.updateSong(): Arg 1 must be a string or null");
			ctx.collect_garbage();
			return undefined;
		}
		// album
		if(!args[2].is_string(ctx) && !args[2].is_null(ctx)){
			critical("Nuvola.updateSong(): Arg 2 must be a string or null");
			ctx.collect_garbage();
			return undefined;
		}
		// album art
		if(!args[3].is_string(ctx) && !args[3].is_null(ctx)){
			critical("Nuvola.updateSong(): Arg 3 must be a string or null");
			ctx.collect_garbage();
			return undefined;
		}
		// state
		if(!args[4].is_string(ctx)){
			critical("Nuvola.updateSong(): Arg 4 must be a string");
			ctx.collect_garbage();
			return undefined;
		}
		
		/* Check if private data is not null */
		var js_api = get_api(self, null);
		if(js_api == null){
			debug("Nuvola.updateSong(): JSApi is null");
			exception = create_exception(ctx, "JSApi is null");
			return undefined;
		}
		else{
			js_api.song_changed(string_or_null(ctx, args[0]),
				string_or_null(ctx, args[1]),
				string_or_null(ctx, args[2]),
				string_or_null(ctx, args[3]),
				string_or_null(ctx, args[4]) ?? STATE_NONE);
		}
		
		ctx.collect_garbage();
		return undefined;
	}
	
	/**
	 * Updates 'sensitivity' of Nuvola Player action. Implementation of a method Nuvola.updateAction().
	 */
	static unowned JS.Value update_action_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception){
		exception = null;
		
		/* Check number of arguments */
		if(args.length != 2){
			critical("Nuvola.updateAction(): Wrong number of arguments: 1 required, %d provided", args.length);
			ctx.collect_garbage();
			return JS.Value.boolean(ctx, false);
		}
		
		/* Check type of arguments */
		// name
		if(!args[0].is_string(ctx)){
			critical("Nuvola.updateAction(): Arg 0 must be a string");
			ctx.collect_garbage();
			return JS.Value.boolean(ctx, false);
		}
		// sensitive
		if(!args[1].is_boolean(ctx)){
			critical("Nuvola.updateAction(): Arg 1 must be a boolean");
			ctx.collect_garbage();
			return JS.Value.boolean(ctx, false);
		}
		
		/* Check if private data is not null */
		var js_api = get_api(self, null);
		if(js_api == null){
			debug("Nuvola.updateAction():  JSApi is null");
			exception = create_exception(ctx, "JSApi is null");
			return JS.Value.boolean(ctx, false);
		}
		
		var name = utf8_string(args[0].to_jsstring(ctx));
		var action = js_api.actions != null ? js_api.actions.get_action(name) : null;
		if(action != null){
			if(!js_api.changed_actions.lookup(name) == null){
				js_api.changed_actions.insert(name, action.sensitive);
			}
			action.sensitive = args[1].to_boolean(ctx);
//~ 			debug("Nuvola.updateAction(): action '%s' = %s", name, action.sensitive.to_string());
		}
		else{
			debug("Nuvola.updateAction(): unknown action '%s'", name);
			ctx.collect_garbage();
			return JS.Value.boolean(ctx, false);
		}
		
		ctx.collect_garbage();
		return JS.Value.boolean(ctx, true);
	}
	
	/**
	 * Default implementation of Nuvola.onMessageReceived callback
	 */
	static unowned JS.Value default_command_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception){
		exception = null;
		string cmd = null;
		if(args.length > 0){
			cmd = string_or_null(ctx, args[0]);
		}
		warning("Called default Nuvola.onMessageReceived() function for command : %s", cmd ?? null);
		ctx.collect_garbage();
		return JS.Value.undefined(ctx);
	}
	
	/**
	 * Triggers Nuvola Player action, implementation of Nuvola.triggerAction()
	 */
	static unowned JS.Value trigger_action_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception){
		exception = null;
		string cmd = null;
		if(args.length > 0){
			cmd = string_or_null(ctx, args[0]);
		}
		/* Check if private data is not null */
		var js_api = get_api(self, null);
		if(cmd == null){
			debug("Nuvola.triggerAction(): Action not specified");
		}
		else if(js_api == null){
			debug("Nuvola.triggerAction(): JSApi is null");
			exception = create_exception(ctx, "JSApi is null");
		}
		else if (js_api.actions != null){
			var action = js_api.actions.get_action(cmd);
			action.activate();
		}
		
		ctx.collect_garbage();
		return JS.Value.undefined(ctx);
	}
	
	/**
	 * Loads service's own configuration to the main object, implementation of Nuvola.loadConfig()
	 */
	static unowned JS.Value load_config_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception){
		exception = null;
		
		/* Check if private data is not null */
		var js_api = get_api(self, null);
		if(js_api == null){
			debug("Nuvola.loadConfig(): JSApi is null");
			exception = create_exception(ctx, "JSApi is null");
		}
		else{
			self.set_property(ctx, new JS.String(CONFIG_OBJECT),
				object_from_JSON(ctx, js_api.config_json));
		}
		ctx.collect_garbage();
		return JS.Value.undefined(ctx);
	}
	
	
	/**
	 * Saves service's own configuration stored in the main object, implementation of Nuvola.saveConfig()
	 */
	static unowned JS.Value save_config_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception){
		exception = null;
		unowned JS.Value undefined = JS.Value.undefined(ctx);
		
		/* Check if private data is not null */
		var js_api = get_api(self, null);
		if(js_api == null){
			debug("Nuvola.saveConfig(): JSApi is null");
			exception = create_exception(ctx, "JSApi is null");
		}
		else{
			unowned JS.Object? config = o_get_object(ctx, self, CONFIG_OBJECT);
			js_api.config_json = config != null ? utf8_string(config.to_JSON(ctx, 4, null)) : ""; 
			js_api.config_changed();
		}
		ctx.collect_garbage();
		return undefined;
	}
	
	/**
	 * Inject JS API to a frame
	 */
	static unowned JS.Value inject_to_frame_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception)
	{
		exception = null;
		string? name = null;
		string? nick = null;
		unowned JS.Value false_val = JS.Value.boolean(ctx, false);
		bool toplevel = true;
		if (args.length == 0)
		{
			exception = create_exception(ctx, "Too few arguments");
			return false_val;
		}
		else if (args.length >= 3 && args[2].is_boolean(ctx))
		{
			toplevel = args[2].to_boolean(ctx);
		}
		
		name = string_or_null(ctx, args[0]);
		if (name == null)
		{
			exception = create_exception(ctx, "Argument 0 must be a non-empty string");
			return false_val;
		}
		
		if (args.length <= 1 || args[1].is_null(ctx) || args[1].is_undefined(ctx))
		{
			nick = name;
		}
		else
		{
			nick = string_or_null(ctx, args[1]);
			if (nick == null)
			{
				exception = create_exception(ctx, "Argument 1 must be null, undefined or a non-empty string");
				return false_val;
			}
		}
		
		/* Check if private data is not null */
		WebKit.WebFrame? frame;
		var js_api = get_api(self, out frame);
		if (js_api == null)
		{
			debug("Nuvola.injectToFrame(): JSApi is null");
			exception = create_exception(ctx, "JSApi is null");
			return false_val;
		}
		
		debug(@"Nuvola.injectToFrame(): $name");
		return JS.Value.boolean(ctx, js_api.inject_to_frame(name, nick, toplevel ? null : frame));
	}
	
	private static JSApi? get_api(JS.Object main_object, out WebKit.WebFrame? frame)
	{
		var bridge = (main_object.get_private() as FrameBridge);
		if (bridge == null)
		{
			frame = null;
			return null;
		}
		else
		{
			frame = bridge.frame;
			return bridge.js_api;
		}
	}
	
	/**
	 * Calls gettext(), implementation of Nuvola.gettext()
	 */
	static unowned JS.Value gettext_func(Context ctx, JS.Object function, JS.Object self, JS.Value[] args, out unowned JS.Value exception){
		exception = null;
		unowned JS.Value undefined = JS.Value.undefined(ctx);
		/* Check number of arguments */
		if(args.length != 1){
			critical("Nuvola.gettext(): Wrong number of arguments: 1 required, %d provided", args.length);
			return undefined;
		}
		
		/* Check type of arguments */
		if(!args[0].is_string(ctx)){
			critical("Nuvola.gettext(): Arg 0 must be a string");
			return undefined;
		}
		
		var value = _(utf8_string(args[0].to_jsstring(ctx)));
		return JS.Value.string(ctx, new JS.String(value));
	}
	
	/**
	 * Default methods of the main object. Other functions are implemented in JavaScript,
	 * see file main.js
	 */
	private static const JS.StaticFunction[] static_functions = {
		{"updateSong", update_song_func, 0},
		{"updateAction", update_action_func, 0},
		{"reloadConfig", load_config_func, 0},
		{"saveConfig", save_config_func, 0},
		{"gettext", gettext_func, 0},
		{"triggerAction", trigger_action_func, 0},
		{"injectToFrame", inject_to_frame_func, 0},
		{MESSAGE_RECEIVER, default_command_func, 0},
		{null, null, 0}
	};
	
	/**
	 * Creates Nuvola Player main object class description
	 */
	private static void create_class(){
		unowned ClassDefinition class_def = {
		    1,                                        
		    JS.ClassAttribute.None,                      
		    "Nuvola JavaScript API",               
		    null,                                       
		    null,            
		    static_functions,         
		    null, 
		    null, 
		    null, 
		    null,
		    null, 
		    null, 
		    null, 
		    null, 
		    null, 
		    null, 
		    null  
		};
		klass = JS.create_class(class_def);
	}
	
	private File config;
	private string config_json{
		owned get{
			try{
				return Diorite.System.read_file(config);
			}
			catch(Error e){
				return "";
			}
		}
		set{
			try{
				Diorite.System.overwrite_file(config, value);
			}
			catch(Error e){
				warning("Unable to save srvice configuration: %s", e.message);
			}
		}
	}
	
	/**
	 * Creates new Nuvola JavaScript API bridge between service and web view.
	 * 
	 * @param storage				storage fon application
	 * @param view					web view whose JavaScript environment will host the API
	 * @param service				service that will be bound to that API
	 * @param actions				Nuvola Player actions
	 * @param late_injection	if true, JSApi will be injected when page is loaded
	 */
	public JSApi(Diorite.Storage storage, Nuvola.WebView view, Service service,
	Diorite.Actions? actions, string main_frame_id, bool late_injection=false)
	{
		this.storage = storage;
		this.service = service;
		this.view = view;
		this.actions = actions;

		var config_dir = service.config_dir;
		config = config_dir.get_parent().get_child(config_dir.get_basename() + ".json");
		integration_file = service.resources.get_child(INTEGRATION_FILENAME);
		changed_actions = new HashTable<string, bool?>(str_hash, str_equal);
		this.alternate_injection = late_injection;
		if (!late_injection)
			view.window_object_cleared.connect(on_window_object_cleared);
		
		view.document_load_finished.connect_after(on_document_loaded);
		view.console_message.connect(console_handler);
		
		frames = new SList<weak WebKit.WebFrame>();
		main_bridge = bridge_frame(main_frame_id, view.get_main_frame());
		
	}
	
	public void deactivate()
	{
		changed_actions.@foreach(reset_action);
		view.window_object_cleared.disconnect(on_window_object_cleared);
		view.document_load_finished.disconnect(on_document_loaded);
		view.console_message.disconnect(console_handler);
	}
	
	~JSApi()
	{
		changed_actions.@foreach(reset_action);
	}
	
	/**
	 * Sets actions to original state.
	 */
	private void reset_action(string name, bool? value)
	{
		var action = actions.get_action(name);
		if (action != null && value != null)
			action.sensitive = value;
	}
	
	/**
	 * JavaScript console handler
	 */
	private bool console_handler(string message, int line_number, string source_id)
	{
		if (view.developer_mode)
			debug("%s:%d: %s", source_id, line_number, message);
		return true;
	}
	
	/**
	 * Called when a web document is loaded.
	 */
	private void on_document_loaded(WebKit.WebFrame frame)
	{
		var bridge = frame.get_data<FrameBridge>(FRAME_PROPERTY_NAME);
		if (bridge == null)
			return;
		debug("document loaded in frame '%s'.", bridge.frame_id);
		var status_code = frame.get_network_response().message.status_code;
		if (status_code >= 400)
		{
			message("Load error: %u in frame '%s'.", status_code, bridge.frame_id);
			/* Just to be sure no JavaScript API method is executed,
			 * because it causes SIGSEGV in JavaScriptCore. */
			bridge.object_this = null;
		}
		else
		{
			if(alternate_injection)
			{
				unowned JS.Context ctx = bridge.get_context();
				inject_js_api(bridge, ctx);
			}
			
			try
			{
				if (bridge.frame_id == PREFERENCES_FRAME_ID)
					bridge.execute_script_from_file(service.resources.get_child(SETTINGS_SCRIPT));
				else
					bridge.execute_script_from_file(integration_file);
			}
			catch(JSError e)
			{
				critical("JSError: %s in frame '%s'.", e.message, bridge.frame_id);
			}
		}
		
		if (frame == view.get_main_frame())
			load_finished();
	}
	
	private void on_window_object_cleared(WebKit.WebFrame frame, void* context, void* window_object)
	{
		var bridge = frame.get_data<FrameBridge>(FRAME_PROPERTY_NAME);
		if (bridge == null)
			return;
		debug("Window object cleared in frame '%s'.", bridge.frame_id);
		unowned JS.Context ctx = (JS.Context) context;
		inject_js_api(bridge, ctx);
	}
	
	/**
	 * Creates the main object and injects it to the JavaScript context
	 * 
	 * @param ctx	hosting JavaScript context
	 * @return true if the main object has been successfully created
	 */
	private bool inject_js_api(FrameBridge bridge, JS.Context ctx)
	{
			bridge.object_this = null;

			if (klass == null)
				create_class();
			unowned JS.Object object_this = ctx.make_object(klass, bridge);
			o_set_string(ctx, object_this, "STATE_PLAYING", STATE_PLAYING);
			o_set_string(ctx, object_this, "STATE_NONE", STATE_NONE);
			o_set_string(ctx, object_this, "STATE_PAUSED", STATE_PAUSED);
			o_set_string(ctx, object_this, "ACTION_PLAY", Actions.PLAY);
			o_set_string(ctx, object_this, "ACTION_PAUSE", Actions.PAUSE);
			o_set_string(ctx, object_this, "ACTION_TOGGLE_PLAY", Actions.TOGGLE_PLAY);
			o_set_string(ctx, object_this, "ACTION_PREV_SONG", Actions.PREV_SONG);
			o_set_string(ctx, object_this, "ACTION_NEXT_SONG",Actions.NEXT_SONG);
			o_set_string(ctx, object_this, "ACTION_THUMBS_UP", Actions.THUMBS_UP);
			o_set_string(ctx, object_this, "ACTION_THUMBS_DOWN", Actions.THUMBS_DOWN);
			o_set_string(ctx, object_this, "ACTION_FAVORITE", Actions.FAVORITE);
			o_set_string(ctx, object_this, "CONFIG_CHANGED", CONFIG_CHANGED_MESSAGE);
			o_set_number(ctx, object_this, "API_VERSION", (double)API_VERSION);
			o_set_number(ctx, object_this, "API_VERSION_MINOR", (double)API_VERSION_MINOR);
			o_set_number(ctx, object_this, "VERSION_MAJOR", (double)Config.VERSION_MAJOR);
			o_set_number(ctx, object_this, "VERSION_MINOR", (double)Config.VERSION_MINOR);
			o_set_number(ctx, object_this, "VERSION_BUGFIX", (double)Config.VERSION_BUGFIX);
			o_set_string(ctx, object_this, "VERSION_SUFFIX", Config.VERSION_SUFFIX);
			o_set_string(ctx, object_this, "MAIN_FRAME_ID", MAIN_FRAME_ID);
			o_set_string(ctx, object_this, "PREFERENCES_FRAME_ID", PREFERENCES_FRAME_ID);
			o_set_string(ctx, object_this, "frame_id", bridge.frame_id);
			object_this.set_property(ctx, new JS.String(CONFIG_OBJECT),
				object_from_JSON(ctx, config_json));
			
			bridge.object_this = object_this;
			
			/* Functions not bound to this.service are written in pure JavaScript.
			 * It is easier than implement them via JavaScriptCore API. */
			var file = storage.get_data_file(MAIN_SCRIPT_FILENAME);
			if (file != null)
			{
				try
				{
					bridge.execute_script_from_file(file);
					if (service.api_major == 1)
					{
						debug("JS API: API 1.x compatibility layer");
						file = storage.get_data_file(COMPAT_SCRIPT_FILENAME);
						if (file != null)
							bridge.execute_script_from_file(file);
					}
					else
					{
						file = storage.get_data_file("js/forms.js");
						if(file != null)
							bridge.execute_script_from_file(file);
					}
				}
				catch(JSError e)
				{
					critical("Unable to execute Nuvola JS API file: %s: %s",
						file.get_uri(), e.message);
					bridge.object_this = null;
					ctx.collect_garbage();
					return false;
				}
				
				ctx.collect_garbage();
				debug("JSAPI injected successfully in frame '%s'", bridge.frame_id);
				api_injected(bridge);
				return true;
			}
			else
			{
				critical("Unable to find Nuvola JS API file.");
				bridge.object_this = null;
				ctx.collect_garbage();
				return false;
			}
	}
	
	private bool inject_to_frame(string name, string nick, WebKit.WebFrame? parent_frame=null)
	{
		var frame = parent_frame ?? view.get_main_frame();
		frame = frame.find_frame(name);
		if (frame == null)
		{
			debug("Frame '%s' not found'", name);
			return false;
		}
		
		var bridge = bridge_frame(nick, frame);
		var res =  inject_js_api(bridge, bridge.get_context());
		if (frame.load_status == WebKit.LoadStatus.FINISHED)
		{
			try
			{
				bridge.execute_script_from_file(integration_file);
			}
			catch(JSError e)
			{
				warning("Failed to exec integration.js: %s", e.message);
				return false;
			}
		}
		return res;
	}
	
	/**
	 * Sends message through JavaScript API
	 * 
	 * @param string	message to send
	 * @throw JSError on failure
	 */
	public void send_message(string message) throws JSError
	{
		foreach (var frame in frames)
		{
			var bridge = frame.get_data<FrameBridge>(FRAME_PROPERTY_NAME);
			assert(bridge != null);
			if (bridge.object_this == null)
				continue; // Bridge seems to be destroyed
			unowned JS.Context ctx = bridge.get_context();
			unowned JS.Object? command_obj = o_get_object(ctx, bridge.object_this, MESSAGE_RECEIVER);
			if (command_obj == null)
			{
				ctx.collect_garbage();
				throw new JSError.NOT_FOUND("MessageReceiver object '%s' not found.'".printf(MESSAGE_RECEIVER));
			}
		
			if(!command_obj.is_function(ctx))
			{
				ctx.collect_garbage();
				throw new JSError.WRONG_TYPE(
					"MessageReceiver object '%s' is not a function.'".printf(MESSAGE_RECEIVER));
			}
		
			void* args[1];
			args[0] = (void*) JS.Value.string(ctx, new JS.String(message));
			JS.Value? exception;
			command_obj.call_as_function(ctx, bridge.object_this, (JS.Value[]) args,  out exception);
			if (exception != null)
			{
				ctx.collect_garbage();
				throw new JSError.FUNC_FAILED(
					"Frame command %s failed in frame '%s': %s".printf(message ?? "(null)",
					bridge.frame_id, string_or_null(ctx, exception) ?? "(null)"));
			}
			ctx.collect_garbage();
		}
	}
	
	private FrameBridge bridge_frame(string frame_id, WebKit.WebFrame frame)
	{
		var bridge = new FrameBridge(frame_id, frame, this);
		frame.set_data<FrameBridge>(FRAME_PROPERTY_NAME, bridge);
		frames.prepend(frame);
		frame.weak_ref(on_frame_finalized);
		bridge_created(bridge);
		return bridge;
	}
	
	private void on_frame_finalized(GLib.Object object)
	{
		var frame = object as WebKit.WebFrame;
		return_if_fail(frame != null);
		frames.remove(frame);
	}
	
	public static bool is_supported(int api_major, int api_minor)
	{
		return (api_major >= COMPAT_API_MAJOR && api_major < API_VERSION) 
		|| (api_major == API_VERSION && api_minor <= API_VERSION_MINOR);
	}
	
	public static bool has_settings(Service service)
	{
		return service.resources.get_child(SETTINGS_SCRIPT).query_exists();
	}	
}

} // namespace Nuvola
